package encapsulation.app;

import java.util.*;

import encapsulation.model.Builder;
import encapsulation.model.Program;

import parser.parsetree.SyntaxTree;
import jst.symtab.Decl;
import jst.symtab.ExecutableCode;
import jst.symtab.FieldDecl;
import jst.symtab.MethodDecl;
import jst.symtab.PackageDecl;
import jst.symtab.ParameterizedClassType;
import jst.symtab.ParameterizedInterfaceType;
import jst.symtab.UserType;
import jst.symtab.link.IdLookup;
import jst.symtab.link.OperationLookup;
import jst.symtab.link.TypeLink;
import jst.symtab.visitor.CompositionVisitor;

/**
 * JSTModelVisitor class, visits the JST model of the program and extracts
 * relevant information to build a simpler model for metrics measurements
 * @author Janina Voigt
 *
 */
public class JSTModelVisitor extends CompositionVisitor {
	
	//constants needed to extract data from parsetrees
	protected static final String expressionName = "Expression";
	protected static final String expressionNameName = "ExpressionName";
	protected static final String expressionStatementName = "ExpressionStatement";
	protected static final String lhsName = "LeftHandSide";
	
	protected Builder builder;	//the builder to build the simpler model of the program
	
	/**
	 * Constructor for JSTModelVisitor objects
	 */
	public JSTModelVisitor(){
		builder = new Builder();
	}
	
	/**
	 * Visit the model of the program using the given visitor
	 */
	public void visitModel(){
		PackageDecl defaultPackage = PackageDecl.getDefaultPackage();
		defaultPackage.accept(this);
	}
	
	/**
	 * Gets the simpler model of the program that was built by the builder
	 * @return
	 */
	public Program getProgram() {
		return builder.getResult();
	}
	
	/**
	 * Visits a user type in the JST model
	 * @param userType the JST UserType to visit
	 */
	public void visitUserType(UserType userType){
		if(userType == null || userType.getSource() == null) {
        	return;
        }
		if(userType instanceof ParameterizedClassType ||
				userType instanceof ParameterizedInterfaceType){
			return;
		}
		builder.addUserType(userType);
		super.visitUserType(userType);
	}

	/**
	 * Visits a field decl in the JST model
	 * @param fieldDecl the JST FieldDecl to visit
	 */
	public void visitFieldDecl(FieldDecl fieldDecl) {
		if(fieldDecl.getOwnerUserType() == null || fieldDecl.getOwnerUserType().getSource() == null) {
        	return;
        }
		if(fieldDecl.getOwnerUserType() instanceof ParameterizedClassType ||
				fieldDecl.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
        super.visitFieldDecl(fieldDecl);
        builder.addField(fieldDecl);
    }
	
	/**
	 * Visits a method decl in the JST model
	 * @param methodDecl the JST MethodDecl to visit
	 */
	public void visitMethodDecl(MethodDecl methodDecl){
		if(methodDecl.getOwnerUserType() == null || methodDecl.getOwnerUserType().getSource() == null) {
        	return;
        }
		if(methodDecl.getOwnerUserType() instanceof ParameterizedClassType ||
				methodDecl.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
		super.visitMethodDecl(methodDecl);
		builder.addMethod(methodDecl);
	}
	
	/**
	 * Visits a block of executable code in the JST model
	 * @param executableCode the JST ExecutableCode to visit
	 */
	public void visitExecutableCode(ExecutableCode executableCode){
		if(executableCode.getOwnerUserType() == null || executableCode.getOwnerUserType().getSource() == null) {
        	return;
        }
		if(executableCode.getOwnerUserType() instanceof ParameterizedClassType ||
				executableCode.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
		super.visitExecutableCode(executableCode);
		
		builder.addBlock(executableCode);
		
		//follow all links (accesses) from the block
		Set<jst.symtab.link.Link> refersTo = executableCode.getReferences();
    	Iterator<jst.symtab.link.Link> refs = refersTo.iterator();
    	
		while (refs.hasNext()) {
			jst.symtab.link.Link link = refs.next();
			
			do {					
				visitAccess(link, executableCode);
				
				if(link instanceof OperationLookup) {
					OperationLookup lookup = (OperationLookup) link;
					
					//for methods call go to the inner links that represent arguments
					for(TypeLink argLink : lookup.getArgLinks()){
						argLink.resolve();
						visitAccess(link, executableCode);
					}
				}
				
				//go to the inner link
				link = getNextLink(link);
			} while(link != null);
		}
	}
	
	/**
	 * Helper method to visit an access from a block to another entity
	 * @param link the link to the accessed entity
	 * @param executableCode the block of code the access occured in
	 */
	private void visitAccess(jst.symtab.link.Link link, ExecutableCode executableCode) {
		Decl decl = link.getDecl();
		
		if(decl == null) {
			return;
		}
		
    	//get the expression used to refer to the decl
		String expression =  getExpression(link, decl);
		if(expression == null) {
			return;
		}
		
		builder.addAccess(decl, expression);
	}
	
	/**
	 * Gets the next inner link given a link
	 * @param link the link that was visited last
	 * @return an inner link of the given link if it exists, null otherwise
	 */
	private jst.symtab.link.Link getNextLink(jst.symtab.link.Link link){
    	if(link instanceof IdLookup){
			return ((IdLookup)link).getQualifier();
		}
		else {
			return null;
		}
    }
	
	/**
     * Gets a string of the part of the expression that's relevant for this access
     * If the decl we are looking for is called a, then:
     * For "a", return a
     * For "this.a", return this.a
     * For "this.b.a", return b.a
     * @param link the link representing the access
     * @param decl the decl being accessed
     * @return the string of code for the given access 
     */
    private String getExpression(jst.symtab.link.Link link, Decl decl) {
    	SyntaxTree source = link.getSource();
    	
    	if(source == null) {
    		return null;
    	}
    	
    	//go up the parsetree until we get to the right node
    	while(source != null &&
    			!source.getSymbolType().getName().equals(expressionName) &&
    			!source.getSymbolType().getName().equals(lhsName) &&
    			!source.getSymbolType().getName().equals(expressionStatementName) &&
    			!source.getSymbolType().getName().equals(expressionNameName)){
    		source = source.getParent();
    	}
    	
    	if(source == null) {
    		return null;
    	}
    	
    	//extract the relevant part of the expression
    	String wholeExpression = " " + source.getValue() + " ";
    	
    	int index = wholeExpression.indexOf(" " + decl.getSimpleName() + " ");
    	if(index == -1) {
    		return null;
    	}
    	int dotIndex = wholeExpression.lastIndexOf(".", index);
    	if(dotIndex < index && dotIndex != -1){
    		//index of the dot before the one immediately before the access we are looking for
    		int previousDot = wholeExpression.lastIndexOf(".", dotIndex - 1);
    		int endIndex = index + decl.getSimpleName().length() + 1;
    		return wholeExpression.substring(previousDot + 1, endIndex).trim();
    	}
    	else {
    		return decl.getSimpleName();
    	}
    }
}
