package parser.grammar;

import java.util.*;

public class Grammar {

    /**
     * Index of names to ensure uniqueness of Symbols.
     * @link aggregation
     *      @associates parser.grammar.Symbol
     * @supplierCardinality 1..*
     * @supplierRole symbols
     */
    protected Map symbols; // String ==> Symbol;

    /**
     * @link aggregation
     *      @associates parser.grammar.Terminal
     * @supplierCardinality 1..*
     * @supplierRole terminals
     */
    protected List terminals;

    /**
     * @link aggregation
     *      @associates parser.grammar.Nonterminal
     * @supplierCardinality 1..*
     * @supplierRole nonterminals
     */
    protected List nonterminals;

    protected int id;

    public Grammar() {
        symbols = new HashMap();
        terminals = new ArrayList();
        nonterminals = new ArrayList();
        id = 1;
    }

    /** Get or construct a terminal symbol. */
    public Terminal getTerminal(String name) {
        Terminal terminal = (Terminal) symbols.get(name);

        if (terminal == null) {
            terminal = new Terminal(name, id++);
            symbols.put(name, terminal);
            terminals.add(terminal);
        }

        return terminal;
    }

    /** Get or construct a nonterminal symbol. */
    public Nonterminal getNonterminal(String name) {
        Nonterminal nonterminal = (Nonterminal) symbols.get(name);

        if (nonterminal == null) {
            nonterminal = new Nonterminal (name, id++);
            symbols.put(name, nonterminal);
            nonterminals.add(nonterminal);

			if (nonterminals.size() == 1)
				nonterminal.setStartSymbol(true);
        }

        return nonterminal;
    }

    /**
     * Get the start symbol for the grammar.
     * The first nonterminal is defined as the start symbol.
     * At least one nonterminal must be defined before calling this method.
     */
    public Nonterminal getStartSymbol() {
        assert nonterminals.size() > 0;

        Nonterminal start = (Nonterminal) nonterminals.get(0);
        assert start.isStartSymbol();
        return start;
    }

    /** Get a symbol for the given name. */
    public Symbol getSymbol(String name) {
        return (Symbol) symbols.get(name);
    }

    public String toString() {
        String result = "";

        Iterator ntIter = nonterminals.iterator();
        while (ntIter.hasNext()) {
            Nonterminal nt = (Nonterminal) ntIter.next();

            Iterator prodIter = nt.getProductions().iterator();
            while (prodIter.hasNext()) {
                result += prodIter.next().toString() + "\n";
            }
        }

        return result;
    }
}
