package parser.grammar;
import java.util.*;

/**
 * Abstract base class for an enum to represent types of symbols in a parse tree.
 * <p>
 * Subclasses separate terminals and nonterminals, because they may be used in
 * different roles, e.g. leaves or branches of parse trees.
 *
 * For a particular grammar, you can generate a subclass of TerminalType
 * that contains lines of the form:
 *     public static final MyLanguageTerminalType INT_TERMINAL = new MyLanguageTerminalType("int");
 * for each terminal.
 *
 * Likewise for nonterminals.
 *
 * Note that the name passed to the SymbolType constructor may be used by client
 * code as the default value of a TerminalSymbol when no other value is supplied.
 * @see TerminalSymbol, NonterminalSymbol.
 */
public abstract class Symbol implements Comparable {

    /** Name of this Symbol. */
    private String name;

    protected int id; // For table driven parsers.

    /**
     * Construct a Symbol with the given name and ID.
     *
     * @param theName describes this symbol.
     * @param theID number of this symbol for table lokups.
     */
    protected Symbol(String theName, int theID) {
        name = theName;
        id = theID;
    }

    public int compareTo(Object o) {
        Symbol that = (Symbol) o;

        return name.compareTo(that.name);
    }

    public String getName() {
        return name;
    }

    /** Is this Symbol a Terminal?  If not, it is a Nonterminal. */
    public abstract boolean isTerminal();

    public Terminal asTerminal() {
        return null;
    }

    public Nonterminal asNonterminal() {
        return null;
    }

    public int getID() {
        return id;
    }

    public String toString() {
        return name;
    }
}
