package parser.pda.stateMachine;

import java.util.*;
import parser.grammar.*;
import parser.pda.tokenStream.*;
import parser.pda.*;
import parser.pda.PdaActor;
import parser.parsetree.Token;

public abstract class State implements Comparable {

    protected Symbol symbol; // May be null (epsilon) for a start state. 

	protected SortedSet items;

    /** @associates parser.pda.stateMachine.State
     * @supplierCardinality 0..*
     * @supplierRole transitions */
    protected Map transitions; // Symbol ==>State

    protected Set reductions; // Productions.

    public State(Symbol theSymbol) {
		symbol = theSymbol;
		items = new TreeSet();
        transitions = new HashMap();
        reductions = new HashSet();
	}

    public State(DottedItem item) {
		this(item.getPriorSymbol());

		addItem(item);
	}

    public State(Set dottedItems) {
		this((DottedItem) dottedItems.iterator().next());

		Iterator itemIter = dottedItems.iterator();
		while (itemIter.hasNext()) {
			DottedItem item = (DottedItem) itemIter.next();

			assert !item.isKernel() || item.getPriorSymbol() == symbol;
			addItem(item);
		}
	}

    public int compareTo(Object o) {
        State that = (State) o;
        return getName().compareTo(that.getName());
    }

	public void addItem(DottedItem item) {
		items.add(item);
	}

	public SortedSet getItems() {
		return items;
	}

	protected SortedSet getKernelItems() {
		SortedSet kernels = new TreeSet();

		Iterator itemIter = items.iterator();
		while(itemIter.hasNext()) {
			DottedItem item = (DottedItem) itemIter.next();

			if (!item.isKernel())
				break; // Kernel items come first.

			kernels.add(item);
		}

		return kernels;
	}

    public String getName() {
        String name = "";

        Iterator itemIter = getKernelItems().iterator();
        while (itemIter.hasNext()) {
            DottedItem item = (DottedItem) itemIter.next();

            name += item.getName();

			if (itemIter.hasNext())
				name += "+";
        }

        return name;
    }

    public Symbol getSymbol() {
        return symbol;
    }

    public State getTransition(Symbol symbol) {
        return (State) transitions.get(symbol);
    }

    public abstract Lookahead getLookahead();

    public void addReduction(Production reduction) {
        reductions.add(reduction);
    }

    public void act(Looker input, PdaActor pda) {
        input.start(); // First look.
        getLookahead().act(input, pda);
    }

    public Set getExpected() {
        return getLookahead().getExpected();
    }

	// Return a String describing the parsing context defined by this state.
	// Used for parser error messages.
	public String getContext() {
		String result = "";

		Iterator itemIter = getKernelItems().iterator();
		while(itemIter.hasNext()) {
			DottedItem item = (DottedItem) itemIter.next();

			if (item.getDotPos() >= item.getProduction().getSize())
				break; // Omit finished items.

			result += "\n" + item;
		}

		return result;
	}

    public String toString() {
        return getName();
    }
}
